from infrasys import Component

from gdm.distribution import DistributionSystem
import pytest

from erad.systems.asset_system import AssetSystem
from erad.constants import ASSET_TYPES


def test_component_addition():
    h = AssetSystem(auto_add_composed_components=True)
    for m in ASSET_TYPES:
        h.add_component(m.example())


def test_component_failure():
    class Testing(Component):
        ...

    test = Testing(name="asdf")

    h = AssetSystem(auto_add_composed_components=True)
    with pytest.raises(AssertionError):
        h.add_component(test)


def test_from_gdm(gdm_system: DistributionSystem):
    asset_system = AssetSystem.from_gdm(gdm_system)
    asset_system.info()


def test_serialization_deserialization(tmp_path):
    h = AssetSystem(auto_add_composed_components=True)
    for m in ASSET_TYPES:
        h.add_component(m.example())
    h.to_json(tmp_path / "asset_system.json")
    AssetSystem.from_json(tmp_path / "asset_system.json")


def test_plot(gdm_system_2: DistributionSystem):
    asset_system = AssetSystem.from_gdm(gdm_system_2)
    asset_system.plot()



from enum import IntEnum


class ScenarioTypes(IntEnum):
    flood_m = 0
    wind_m_per_s = 1
    fire_m = 2
    earthquake_pga = 3


class NodeTypes(IntEnum):
    transmission_tower = 5
    distribution_poles = 6
    transmission_junction_box = 11
    distribution_junction_box = 12


class AssetTypes(IntEnum):
    substation = 0
    solar_panels = 1
    distribution_underground_cables = 2
    transmission_underground_cables = 3
    battery_storage = 4
    transmission_tower = 5
    distribution_poles = 6
    transmission_overhead_lines = 7
    distribution_overhead_lines = 8
    transformer_mad_mount = 9
    transformer_pole_mount = 10
    transmission_junction_box = 11
    distribution_junction_box = 12
    switch = 13

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_

    @classmethod
    def has_asset(cls, asset):
        return asset in cls.__members__



import gdm.distribution.components as gdc
import gdm.distribution.equipment as gde

from erad.enums import AssetTypes
from erad.models.asset_mapping import ComponentFilterModel


asset_to_gdm_mapping = {
    AssetTypes.switch: [
        ComponentFilterModel(
            component_type=gdc.DistributionSwitchBase,
        ),
    ],
    AssetTypes.switch: [
        ComponentFilterModel(
            component_type=gdc.DistributionSwitchBase,
        ),
    ],
    AssetTypes.substation: [
        ComponentFilterModel(
            component_type=gdc.DistributionVoltageSource,
        ),
    ],
    AssetTypes.solar_panels: [
        ComponentFilterModel(
            component_type=gdc.DistributionSolar,
        ),
    ],
    AssetTypes.battery_storage: [
        ComponentFilterModel(
            component_type=gdc.DistributionBattery,
        ),
    ],
    AssetTypes.distribution_underground_cables: [
        ComponentFilterModel(
            component_type=gdc.GeometryBranch,
            component_filter=lambda x: x.equipment.conductors[0].__class__
            == gde.ConcentricCableEquipment
            and x.buses[0].rated_voltage.to("kilovolt").magnitude < 35.0,
        ),
        ComponentFilterModel(
            component_type=gdc.MatrixImpedanceBranch,
            component_filter=lambda x: x.equipment.c_matrix[0, 0]
            .to("microfarad/kilometer")
            .magnitude
            > 0.05
            and x.buses[0].rated_voltage.to("kilovolt").magnitude < 35.0,
        ),
    ],
    AssetTypes.distribution_overhead_lines: [
        ComponentFilterModel(
            component_type=gdc.GeometryBranch,
            component_filter=lambda x: (
                x.equipment.conductors[0].__class__ == gde.BareConductorEquipment
            )
            and x.buses[0].rated_voltage.to("kilovolt").magnitude < 35.0,
        ),
        ComponentFilterModel(
            component_type=gdc.MatrixImpedanceBranch,
            component_filter=lambda x: x.equipment.c_matrix[0, 0]
            .to("microfarad/kilometer")
            .magnitude
            < 0.05
            and x.buses[0].rated_voltage.to("kilovolt").magnitude < 35.0,
        ),
    ],
    AssetTypes.transmission_underground_cables: [
        ComponentFilterModel(
            component_type=gdc.GeometryBranch,
            component_filter=lambda x: x.equipment.conductors[0].__class__
            == gde.ConcentricCableEquipment
            and x.buses[0].rated_voltage.to("kilovolt").magnitude > 35.0,
        ),
    ],
    AssetTypes.transmission_overhead_lines: [
        ComponentFilterModel(
            component_type=gdc.GeometryBranch,
            component_filter=lambda x: x.equipment.conductors[0].__class__
            == gde.BareConductorEquipment
            and x.buses[0].rated_voltage.to("kilovolt").magnitude > 35.0,
        ),
    ],
}



import plotly.graph_objects as go
from pydantic import ConfigDict
from infrasys import Component


class BaseDisasterModel(Component):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    def plot(
        self, time_index: int, figure: go.Figure, map_obj: go.Scattergeo | go.Scattermap
    ) -> int:
        raise NotImplementedError("This method should be implemented in the derived classes")



from datetime import datetime
import sqlite3
import os

from pydantic import field_serializer, field_validator
from infrasys.quantities import Distance
from shapely.geometry import Point
import plotly.graph_objects as go
import geopandas as gpd
import pandas as pd

from erad.models.hazard.common import ERAD_DB, HISTROIC_HURRICANE_TABLE
from erad.models.hazard.base_models import BaseDisasterModel
from erad.quantities import Speed, Pressure


class WindModel(BaseDisasterModel):
    timestamp: datetime
    center: Point
    max_wind_speed: Speed
    radius_of_max_wind: Distance
    radius_of_closest_isobar: Distance
    air_pressure: Pressure

    @field_validator("center", mode="before")
    def deserialize_point(cls, value):
        if isinstance(value, dict) and value.get("type") == "Point":
            coords = value["coordinates"]
            return Point(coords)
        return value

    @field_serializer("center")
    def serialize_location(self, point: Point, _info):
        return {"type": "Point", "coordinates": (point.x, point.y)}

    @classmethod
    def example(cls) -> "WindModel":
        return WindModel(
            name="hurricane 1",
            timestamp=datetime.now(),
            center=Point(-121.93036, 36.60144),
            max_wind_speed=Speed(50, "miles/hour"),
            air_pressure=Pressure(1013.25, "hPa"),
            radius_of_max_wind=Distance(50, "miles"),
            radius_of_closest_isobar=Distance(300, "miles"),
        )

    @classmethod
    def from_hurricane_sid(cls, hurricane_sid: str) -> list["WindModel"]:
        assert os.path.exists(ERAD_DB), f"The data file {ERAD_DB} not found"
        conn = sqlite3.connect(ERAD_DB)
        hurricane_data = pd.read_sql(
            f"SELECT * FROM {HISTROIC_HURRICANE_TABLE} WHERE `SID ` = '{hurricane_sid}';", conn
        )
        cols = [
            "LAT (degrees_north)",
            "LON (degrees_east)",
            "USA_WIND (kts)",
            "USA_ROCI (nmile)",
            "USA_RMW (nmile)",
            "USA_POCI (mb)",
            "ISO_TIME ",
        ]
        hurricane_data = hurricane_data[cols]
        for col in cols:
            hurricane_data = hurricane_data[hurricane_data[col] != " "]
        if hurricane_data.empty:
            raise ValueError(
                f"Hurricane '{hurricane_sid}'  not found in column 'SID', table '{HISTROIC_HURRICANE_TABLE}' in the database"
            )
        conn.close()
        geometry = [
            Point(lat, lon)
            for lat, lon in zip(
                hurricane_data["LAT (degrees_north)"], hurricane_data["LON (degrees_east)"]
            )
        ]
        hurricane_data["ISO_TIME "] = pd.to_datetime(hurricane_data["ISO_TIME "])
        hurricane_data = gpd.GeoDataFrame(hurricane_data, geometry=geometry)
        hurricane_data.set_crs("epsg:4326")
        track = []
        for idx, row in hurricane_data.iterrows():
            track.append(
                WindModel(
                    name=hurricane_sid,
                    timestamp=row["ISO_TIME "],
                    center=row["geometry"],
                    max_wind_speed=Speed(float(row["USA_WIND (kts)"]), "knots"),
                    radius_of_max_wind=Distance(float(row["USA_RMW (nmile)"]), "nautical_mile"),
                    radius_of_closest_isobar=Distance(
                        float(row["USA_ROCI (nmile)"]), "nautical_mile"
                    ),
                    air_pressure=Pressure(float(row["USA_POCI (mb)"]), "millibar"),
                )
            )

        return track

    def plot(
        self,
        time_index: int = 0,
        figure: go.Figure = go.Figure(),
        map_obj: type[go.Scattergeo | go.Scattermap] = go.Scattermap,
    ) -> int:
        figure.add_trace(
            map_obj(
                lat=[self.center.x],
                lon=[self.center.y],
                mode="markers",
                marker=dict(
                    size=[self.radius_of_closest_isobar.magnitude / 5],
                    color="lightblue",
                    opacity=0.4,
                ),
                name="Radius of closest isobar",  # Name for the legend
                visible=(time_index == 0),
            )
        )

        figure.add_trace(
            map_obj(
                lat=[self.center.x],
                lon=[self.center.y],
                mode="markers",
                marker=dict(
                    size=[self.radius_of_max_wind.magnitude / 5],
                    color=[self.max_wind_speed.magnitude],
                    showscale=False,
                    opacity=0.4,
                ),
                visible=(time_index == 0),
                hovertext=[
                    f"""
                    <br> <b>Max wind speed:</b> {self.max_wind_speed}
                    <br> <b>Radius of max wind speed:</b> {self.radius_of_max_wind}
                    <br> <b>Radius of closest isobar:</b> {self.radius_of_closest_isobar}
                    <br> <b>Air pressure:</b> {self.air_pressure}
                    """
                ],
                name="Radius of max wind",  # Name for the legend
            )
        )
        return 2



from infrasys import System

from gdm.distribution.enums import PlotingStyle, MapType
import plotly.graph_objects as go

from erad.default_fragility_curves import DEFAULT_FRAGILTY_CURVES
from erad.constants import HAZARD_MODELS
import erad.models.fragility_curve as fc
import erad.models.hazard as hz


class HazardSystem(System):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def add_component(self, component, **kwargs):
        assert isinstance(
            component, HAZARD_MODELS
        ), f"Unsupported model type {component.__class__.__name__}"
        return super().add_component(component, **kwargs)

    def add_components(self, *components, **kwargs):
        assert all(isinstance(component, HAZARD_MODELS) for component in components), (
            "Unsupported model types in passed component. Valid types are: \n"
            + "\n".join([s.__name__ for s in HAZARD_MODELS])
            + "\nPassed components: \n"
            + "\n".join(set([component.__class__.__name__ for component in components]))
        )
        return super().add_components(*components, **kwargs)

    def to_json(self, filename, overwrite=False, indent=None, data=None):
        if not list(self.get_components(fc.HazardFragilityCurves)):
            self.add_components(*DEFAULT_FRAGILTY_CURVES)
        return super().to_json(filename, overwrite, indent, data)

    @classmethod
    def fire_example(cls) -> "HazardSystem":
        hazard = hz.FireModel.example()
        system = HazardSystem(auto_add_composed_components=True)
        system.add_component(hazard)
        return system

    @classmethod
    def wind_example(cls) -> "HazardSystem":
        hazard = hz.WindModel.example()
        system = HazardSystem(auto_add_composed_components=True)
        system.add_component(hazard)
        return system

    @classmethod
    def earthquake_example(cls) -> "HazardSystem":
        hazard = hz.EarthQuakeModel.example()
        system = HazardSystem(auto_add_composed_components=True)
        system.add_component(hazard)
        return system

    @classmethod
    def flood_example(cls) -> "HazardSystem":
        hazard = hz.FloodModel.example()
        system = HazardSystem(auto_add_composed_components=True)
        system.add_component(hazard)
        return system

    @classmethod
    def multihazard_example(cls) -> "HazardSystem":
        wind_hazard = hz.WindModel.example()
        flood_hazard = hz.FloodModel.example()
        system = HazardSystem(auto_add_composed_components=True)
        system.add_component(wind_hazard)
        system.add_component(flood_hazard)
        return system

    def plot(
        self,
        show: bool = True,
        show_legend: bool = True,
        map_type: MapType = MapType.SCATTER_MAP,
        style: PlotingStyle = PlotingStyle.CARTO_POSITRON,
        zoom_level: int = 11,
        figure=go.Figure(),
    ):
        timestamps = sorted(
            [model.timestamp for model in self.get_components(hz.BaseDisasterModel)]
        )

        steps = []
        for i, ts in enumerate(timestamps):
            hazards = self.get_components(
                hz.BaseDisasterModel, filter_func=lambda x: x.timestamp == ts
            )
            map_obj = getattr(go, map_type.value)
            num_traces = 0
            for hazard in hazards:
                num_traces += hazard.plot(i, figure, map_obj)
            vis = [j == i for j in range(len(timestamps))]
            result = [x for x in vis for _ in range(num_traces)]
            steps.append(dict(method="update", label=str(ts), args=[{"visible": result}]))

        sliders = [dict(active=0, pad={"t": 50}, steps=steps)]

        if map_type == MapType.SCATTER_MAP:
            figure.update_layout(
                map={
                    "style": style.value,
                    "zoom": zoom_level,
                },
                sliders=sliders,
                showlegend=True if show_legend else False,
            )
        else:
            figure.update_layout(
                geo=dict(
                    projection_scale=zoom_level,
                ),
                sliders=sliders,
                showlegend=True if show_legend else False,
            )

        if show:
            figure.show()

        return figure



import erad.models.fragility_curve as frag
from erad.enums import AssetTypes
from erad.quantities import Speed


DEFAULT_WIND_SPEED_FRAGILITY_CURVES = frag.HazardFragilityCurves(
    asset_state_param="wind_speed",
    curves=[
        frag.FragilityCurve(
            asset_type=AssetTypes.switch,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.45, "m/s"), Speed(50, "m/s"), 1 / 0.45]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.battery_storage,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.45, "m/s"), Speed(50, "m/s"), 1 / 0.45]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_junction_box,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.4, "m/s"), Speed(50, "m/s"), 1 / 0.4]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_overhead_lines,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.35, "m/s"), Speed(45, "m/s"), 1 / 0.35]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_poles,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.4, "m/s"), Speed(45, "m/s"), 1 / 0.4]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_underground_cables,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.4, "m/s"), Speed(55, "m/s"), 1 / 0.4]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.solar_panels,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.3, "m/s"), Speed(55, "m/s"), 1 / 0.3]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.substation,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.4, "m/s"), Speed(55, "m/s"), 1 / 0.40]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transformer_mad_mount,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.35, "m/s"), Speed(50, "m/s"), 1 / 0.35]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transformer_pole_mount,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.3, "m/s"), Speed(45, "m/s"), 1 / 0.30]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_junction_box,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.54, "m/s"), Speed(55, "m/s"), 1 / 0.54]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_overhead_lines,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.35, "m/s"), Speed(50, "m/s"), 1 / 0.35]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_tower,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.4, "m/s"), Speed(55, "m/s"), 1 / 0.40]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_underground_cables,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.4, "m/s"), Speed(60, "m/s"), 1 / 0.40]
            ),
        ),
    ],
)



from gdm.quantities import Distance

import erad.models.fragility_curve as frag
from erad.enums import AssetTypes

DEFAULT_FIRE_BOUNDARY_FRAGILITY_CURVES = frag.HazardFragilityCurves(
    asset_state_param="fire_boundary_dist",
    curves=[
        frag.FragilityCurve(
            asset_type=AssetTypes.switch,
            prob_function=frag.ProbabilityFunction(
                distribution="expon", parameters=[Distance(0.65, "km"), 0.95]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.battery_storage,
            prob_function=frag.ProbabilityFunction(
                distribution="expon", parameters=[Distance(0.65, "km"), 0.95]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_junction_box,
            prob_function=frag.ProbabilityFunction(
                distribution="expon", parameters=[Distance(0.5, "km"), 0.95]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_overhead_lines,
            prob_function=frag.ProbabilityFunction(
                distribution="expon", parameters=[Distance(0.5, "km"), 0.95]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_poles,
            prob_function=frag.ProbabilityFunction(
                distribution="expon", parameters=[Distance(1.0, "km"), 0.95]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_underground_cables,
            prob_function=frag.ProbabilityFunction(
                distribution="expon", parameters=[Distance(0.1, "km"), 0.95]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.solar_panels,
            prob_function=frag.ProbabilityFunction(
                distribution="expon", parameters=[Distance(0.55, "km"), 0.95]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.substation,
            prob_function=frag.ProbabilityFunction(
                distribution="expon", parameters=[Distance(0.7, "km"), 0.95]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transformer_mad_mount,
            prob_function=frag.ProbabilityFunction(
                distribution="expon", parameters=[Distance(0.9, "km"), 0.95]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transformer_pole_mount,
            prob_function=frag.ProbabilityFunction(
                distribution="expon", parameters=[Distance(1.0, "km"), 0.95]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_junction_box,
            prob_function=frag.ProbabilityFunction(
                distribution="expon", parameters=[Distance(0.55, "km"), 0.95]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_overhead_lines,
            prob_function=frag.ProbabilityFunction(
                distribution="expon", parameters=[Distance(1.1, "km"), 0.95]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_tower,
            prob_function=frag.ProbabilityFunction(
                distribution="expon", parameters=[Distance(0.7, "km"), 0.95]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_underground_cables,
            prob_function=frag.ProbabilityFunction(
                distribution="expon", parameters=[Distance(0.15, "km"), 0.95]
            ),
        ),
    ],
)



"""Module for managing constants in ERAD package.

_Do not change this constants in your code._
"""

from datetime import datetime
from pathlib import Path

import elevation

import erad.models.fragility_curve as frag
import erad.models.probability as prob
import erad.models.hazard as hazard
import erad.models.asset as asset

from erad.enums import AssetTypes

# Get list of continuous distributions

RASTER_DOWNLOAD_PATH = Path(elevation.CACHE_DIR) / "SRTM1" / "raster.tif"
ROOT_PATH = Path(__file__).parent.parent.parent
TEST_PATH = Path(__file__).parent.parent.parent / "tests"
DATA_FOLDER_NAME = "data"
DATA_FOLDER = TEST_PATH / DATA_FOLDER_NAME

DEFAULT_TIME_STAMP = datetime(1970, 1, 1, 0, 0, 0)

DEFAULT_HEIGHTS_M = {
    AssetTypes.substation: 0.0,
    AssetTypes.solar_panels: 3.0,
    AssetTypes.distribution_underground_cables: -1.0,
    AssetTypes.transmission_underground_cables: -1.0,
    AssetTypes.battery_storage: 1.0,
    AssetTypes.transmission_tower: 0.0,
    AssetTypes.distribution_poles: 0.0,
    AssetTypes.transmission_overhead_lines: 30.0,
    AssetTypes.distribution_overhead_lines: 10.0,
    AssetTypes.transformer_mad_mount: 0.3,
    AssetTypes.transformer_pole_mount: 6.0,
    AssetTypes.transmission_junction_box: -1.0,
    AssetTypes.distribution_junction_box: -1.0,
    AssetTypes.switch: 6.0,
}

ASSET_TYPES = (
    asset.AssetState,
    asset.Asset,
    prob.AccelerationProbability,
    prob.TemperatureProbability,
    prob.DistanceProbability,
    prob.SpeedProbability,
)

HAZARD_TYPES = (
    hazard.EarthQuakeModel,
    hazard.FloodModel,
    hazard.FireModel,
    hazard.WindModel,
)

HAZARD_MODELS = (
    hazard.EarthQuakeModel,
    hazard.FloodModelArea,
    hazard.FireModelArea,
    hazard.FloodModel,
    hazard.FireModel,
    hazard.WindModel,
    frag.ProbabilityFunction,
    frag.FragilityCurve,
    frag.HazardFragilityCurves,
)

SUPPORTED_MODELS = [
    hazard.EarthQuakeModel,
    hazard.FloodModelArea,
    hazard.FireModelArea,
    hazard.FloodModel,
    hazard.FireModel,
    hazard.WindModel,
    prob.AccelerationProbability,
    prob.TemperatureProbability,
    prob.DistanceProbability,
    prob.SpeedProbability,
    frag.ProbabilityFunction,
    frag.FragilityCurve,
    frag.HazardFragilityCurves,
    asset.AssetState,
    asset.Asset,
]



from typing import Annotated

from gdm.quantities import Distance
from infrasys import Component
from pydantic import Field
from abc import ABC

from erad.quantities import Speed, Acceleration, Temperature


class BaseProbabilityModel(Component, ABC):
    name: str = ""
    survival_probability: Annotated[
        float,
        Field(1.0, ge=0, le=1, description="Asset survival probability"),
    ]


class SpeedProbability(BaseProbabilityModel):
    speed: Annotated[
        Speed,
        Field(
            ...,
            description="Represents the speed of a scenario parameter experienced by the asset e.g speed of wind",
        ),
    ]

    @classmethod
    def example(cls) -> "SpeedProbability":
        return SpeedProbability(
            speed=Speed(50, "m/s"),
            survival_probability=1.0,
        )


class TemperatureProbability(BaseProbabilityModel):
    temperature: Annotated[
        Temperature,
        Field(..., description="Temperature of the asset"),
    ]

    @classmethod
    def example(cls) -> "TemperatureProbability":
        return TemperatureProbability(
            temperature=Temperature(0, "degC"),
            survival_probability=1.0,
        )


class DistanceProbability(BaseProbabilityModel):
    distance: Annotated[
        Distance,
        Field(
            Distance(-9999, "m"),
            description="Distance of asset from the source / boundary of a disaster event",
        ),
    ]

    @classmethod
    def example(cls) -> "DistanceProbability":
        return DistanceProbability(
            distance=Distance(0, "m"),
            survival_probability=1.0,
        )


class AccelerationProbability(BaseProbabilityModel):
    name: str = ""
    acceleration: Acceleration = Acceleration(0, "m/s**2")

    @classmethod
    def example(cls) -> "AccelerationProbability":
        return AccelerationProbability(
            acceleration=Acceleration(0, "m/s**2"),
            survival_probability=1.0,
        )



import erad.models.fragility_curve as frag
from erad.quantities import Acceleration
from erad.enums import AssetTypes

DEFAULT_PEAK_GROUND_ACCELERATION_FRAGILITY_CURVES = frag.HazardFragilityCurves(
    asset_state_param="peak_ground_acceleration",
    curves=[
        frag.FragilityCurve(
            asset_type=AssetTypes.switch,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[
                    Acceleration(0.8, "m/s**2"),
                    Acceleration(0.40 * 9.81, "m/s**2"),
                    1 / 0.80,
                ],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.battery_storage,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[
                    Acceleration(0.8, "m/s**2"),
                    Acceleration(0.40 * 9.81, "m/s**2"),
                    1 / 0.80,
                ],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_junction_box,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[
                    Acceleration(0.5, "m/s**2"),
                    Acceleration(0.45 * 9.81, "m/s**2"),
                    1 / 0.50,
                ],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_overhead_lines,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[
                    Acceleration(0.55, "m/s**2"),
                    Acceleration(0.50 * 9.81, "m/s**2"),
                    1 / 0.55,
                ],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_poles,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[
                    Acceleration(0.6, "m/s**2"),
                    Acceleration(0.40 * 9.81, "m/s**2"),
                    1 / 0.6,
                ],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_underground_cables,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[
                    Acceleration(0.8, "m/s**2"),
                    Acceleration(1.0 * 9.81, "m/s**2"),
                    1 / 0.8,
                ],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.solar_panels,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[
                    Acceleration(0.45, "m/s**2"),
                    Acceleration(0.45 * 9.81, "m/s**2"),
                    1 / 0.45,
                ],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.substation,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[
                    Acceleration(0.5, "m/s**2"),
                    Acceleration(0.40 * 9.81, "m/s**2"),
                    1 / 0.5,
                ],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transformer_mad_mount,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[
                    Acceleration(0.6, "m/s**2"),
                    Acceleration(0.40 * 9.81, "m/s**2"),
                    1 / 0.6,
                ],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transformer_pole_mount,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[
                    Acceleration(0.8, "m/s**2"),
                    Acceleration(0.50 * 9.81, "m/s**2"),
                    1 / 0.70,
                ],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_junction_box,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[
                    Acceleration(0.5, "m/s**2"),
                    Acceleration(0.50 * 9.81, "m/s**2"),
                    1 / 0.50,
                ],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_overhead_lines,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[
                    Acceleration(0.6, "m/s**2"),
                    Acceleration(0.60 * 9.81, "m/s**2"),
                    1 / 0.60,
                ],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_tower,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[
                    Acceleration(0.4, "m/s**2"),
                    Acceleration(0.35 * 9.81, "m/s**2"),
                    1 / 0.40,
                ],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_underground_cables,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[
                    Acceleration(0.55, "m/s**2"),
                    Acceleration(0.80 * 9.81, "m/s**2"),
                    1 / 0.55,
                ],
            ),
        ),
    ],
)



from functools import cached_property
from typing import Literal
from pathlib import Path


from infrasys import Component, BaseQuantity
from scipy.stats import _continuous_distns
from pydantic import field_validator
import plotly.express as px
import numpy as np

from erad.probability_builder import ProbabilityFunctionBuilder
from erad.models.asset import AssetState
from erad.enums import AssetTypes
from erad.quantities import Speed


FRAGILITY_CURVE_TYPES = [
    fc for fc in AssetState.model_fields if fc not in ["name", "uuid", "timestamp"]
]
SUPPORTED_CONT_DIST = [name for name in dir(_continuous_distns) if not name.startswith("_")]


class ProbabilityFunction(Component):
    name: str = ""
    distribution: Literal[*SUPPORTED_CONT_DIST]
    parameters: list[float | BaseQuantity]

    @field_validator("parameters")
    def validate_parameters(cls, value):
        if not any(isinstance(v, BaseQuantity) for v in value):
            raise ValueError("There should be atleast one BaseQuantity in the parameters")

        units = set([v.units for v in value if isinstance(v, BaseQuantity)])
        if not len(units) == 1:
            raise ValueError("All BaseQuantities should have the same units")
        return value

    @cached_property
    def prob_model(self) -> ProbabilityFunctionBuilder:
        return ProbabilityFunctionBuilder(self.distribution, self.parameters)

    @classmethod
    def example(cls) -> "ProbabilityFunction":
        return ProbabilityFunction(
            distribution="norm",
            parameters=[Speed(1.5, "m/s"), 2],
        )


class FragilityCurve(Component):
    name: str = ""
    asset_type: AssetTypes
    prob_function: ProbabilityFunction

    @classmethod
    def example(cls) -> "FragilityCurve":
        return FragilityCurve(
            asset_type=AssetTypes.substation,
            prob_function=ProbabilityFunction.example(),
        )


class HazardFragilityCurves(Component):
    name: str = "DEFAULT_CURVES"
    asset_state_param: Literal[*FRAGILITY_CURVE_TYPES]
    curves: list[FragilityCurve]

    @classmethod
    def example(cls) -> "HazardFragilityCurves":
        return HazardFragilityCurves(
            asset_state_param="peak_ground_acceleration",
            curves=[FragilityCurve.example()],
        )

    def plot(
        self, file_path: Path, x_min: float = 0, x_max: float = None, number_of_points: int = 100
    ):
        """Plot the fragility curves."""
        file_path = Path(file_path)
        assert file_path.suffix.lower() == ".html", "File path should be an HTML file"

        if not self.curves:
            raise ValueError("No curves to plot")

        quantities = [
            p for p in self.curves[0].prob_function.parameters if isinstance(p, BaseQuantity)
        ]

        x_label = self.asset_state_param.replace("_", " ").title()
        x = np.linspace(x_min, x_max, number_of_points)

        plot_data = {"x": [], "y": [], "Asset Type": []}

        for curve in self.curves:
            model_class = quantities[0].__class__
            units = quantities[0].units
            prob_model = curve.prob_function.prob_model
            y = prob_model.probability(model_class(x, units))
            label = curve.asset_type.name.replace("_", " ").title()

            plot_data["x"].extend(x)
            plot_data["y"].extend(y)
            plot_data["Asset Type"].extend([label] * len(x))

        fig = px.line(
            plot_data,
            x="x",
            y="y",
            color="Asset Type",
            labels={"x": f"{x_label} [{units}]", "y": "Probability of Failure"},
            title=f"Fragility Curves for {x_label}",
        )

        fig.show()
        fig.write_html(file_path)



from datetime import datetime
from uuid import UUID
import math

from pydantic import computed_field, Field
from infrasys.quantities import Distance
from geopy.distance import geodesic
from shapely.geometry import Point
from pyhigh import get_elevation
from infrasys import Component
from loguru import logger

# from erad.constants import RASTER_DOWNLOAD_PATH
from erad.quantities import Acceleration, Speed
from erad.models.probability import (
    AccelerationProbability,
    DistanceProbability,
    SpeedProbability,
)
from erad.enums import AssetTypes
import erad.models.hazard as hz


class AssetState(Component):
    name: str = ""
    timestamp: datetime
    wind_speed: SpeedProbability | None = None
    flood_velocity: SpeedProbability | None = None
    flood_depth: DistanceProbability | None = None
    fire_boundary_dist: DistanceProbability | None = None
    peak_ground_velocity: SpeedProbability | None = None
    peak_ground_acceleration: AccelerationProbability | None = None

    def calculate_earthquake_vectors(
        self, asset_coordinate: Point, hazard_model: hz.EarthQuakeModel
    ):
        """
        Sources:

        Atkinson, G.M., & Wald, D.J. (2007). “Did You Feel It?” intensity data: A surprisingly good
        measure of earthquake ground motion. Seismological Research Letters, 78(3), 362–368.
        """

        epicenter_distance = Distance(
            geodesic(
                (hazard_model.origin.y, hazard_model.origin.x),
                (asset_coordinate.y, asset_coordinate.x),
            ).km,
            "km",
        )

        hypocentral_distance = (hazard_model.depth**2 + epicenter_distance**2) ** 0.5

        mmi = (
            3.31
            + 1.28 * hazard_model.magnitude
            - 1.42 * math.log10(hypocentral_distance.to("km").magnitude)
        )  # Modified Mercalli Intensity

        log_pgv = (mmi - 3.78) / 1.47
        pgv = 10**log_pgv  # cm/s
        self.peak_ground_velocity = SpeedProbability(
            speed=Speed(pgv, "centimeter/second"),
            survival_probability=1,
        )

        log_pga = (mmi - 1.78) / 3.70
        pga_per_g = 10**log_pga  # % of g
        self.peak_ground_acceleration = AccelerationProbability(
            acceleration=Acceleration(pga_per_g / 100.0 * 9.80665, "meter/second**2"),
            survival_probability=1,
        )

    def calculate_fire_vectors(self, asset_coordinate: Point, hazard_model: hz.FireModel):
        """
        Sources:

        “Integrating Wildfire Risk Modeling with Electric Grid Asset Management” (EPRI)
        “Modeling and Mitigating Wildfire Risk to the Power Grid” (PNNL, 2021)
        “A Framework for Assessing Wildfire Ignition Risk on Distribution Systems”
            (IEEE Transactions on Power Systems, 2020)
        Improving Grid Safety and Resilience to Mitigate Ignition Incident and Fire Risks – EPRI
        Wildfire Risk Reduction Methods 2024 – EPRI
        """

        distances_to_wild_fire_boundaries = []
        for wild_fire_area in hazard_model.affected_areas:
            if wild_fire_area.affected_area.contains(asset_coordinate):
                distance = 0
            else:
                distance = wild_fire_area.affected_area.exterior.distance(asset_coordinate)
            distances_to_wild_fire_boundaries.append(distance)
        minimum_distance_km = min(distances_to_wild_fire_boundaries)
        self.fire_boundary_dist = DistanceProbability(
            distance=Distance(minimum_distance_km, "kilometer"),
            survival_probability=1,
        )

        # survival_probability = 1 - 0.95 * math.exp(-k * minimum_distance_km)

    def calculate_wind_vectors(self, asset_coordinate: Point, hazard_model: hz.WindModel):
        r = Distance(
            geodesic(
                (hazard_model.center.y, hazard_model.center.x),
                (asset_coordinate.y, asset_coordinate.x),
            ).km,
            "km",
        )

        r = r.to("nautical_mile")
        r_max = hazard_model.radius_of_closest_isobar.to("nautical_mile")
        r_v_max = hazard_model.radius_of_max_wind.to("nautical_mile")
        v_max = hazard_model.max_wind_speed.to("knot")

        k = 1.14
        b = 10

        a = math.log(b) / (r_max - r_v_max)
        m = (1 / r_v_max) * math.log(k / (k + 1))

        if r >= 0 and r < r_v_max:
            wind_speed = k * v_max * (1 - math.exp(-m * r))
        elif r >= r_v_max and r < r_max:
            wind_speed = v_max * math.exp(-a * (r - r_v_max))
        else:
            wind_speed = 0

        self.wind_speed = SpeedProbability(
            speed=wind_speed.to("miles/hour"),
            survival_probability=1,
        )

    def calculate_flood_vectors(
        self, asset_coordinate: Point, hazard_model: hz.FloodModel, asset_total_elevation: Distance
    ):
        for area in hazard_model.affected_areas:
            if area.affected_area.contains(asset_coordinate):
                self.flood_depth = DistanceProbability(
                    distance=area.water_elevation - asset_total_elevation,
                    survival_probability=1,
                )
                self.flood_velocity = SpeedProbability(
                    speed=area.water_velocity,
                    survival_probability=1,
                )

        if self.flood_depth is None and self.flood_velocity is None:
            self.flood_depth = DistanceProbability(
                distance=Distance(-9999, "meter"),
                survival_probability=1,
            )
            self.flood_velocity = SpeedProbability(
                speed=Speed(0, "meter / second"),
                survival_probability=1,
            )

    @computed_field
    @property
    def survival_probability(self) -> float:
        wind_survival_probability = self.wind_speed.survival_probability if self.wind_speed else 1
        water_flow_survival_probability = (
            self.flood_velocity.survival_probability if self.flood_velocity else 1
        )
        water_level_survivial_probability = (
            self.flood_depth.survival_probability if self.flood_depth else 1
        )
        fire_survival_probability = (
            self.fire_boundary_dist.survival_probability if self.fire_boundary_dist else 1
        )
        pgv_survival_probability = (
            self.peak_ground_velocity.survival_probability if self.peak_ground_velocity else 1
        )
        pga_survival_probability = (
            self.peak_ground_acceleration.survival_probability
            if self.peak_ground_acceleration
            else 1
        )

        return (
            wind_survival_probability
            * water_flow_survival_probability
            * water_level_survivial_probability
            * fire_survival_probability
            * pgv_survival_probability
            * pga_survival_probability
        )

    @classmethod
    def example(cls) -> "AssetState":
        return AssetState(
            timestamp=datetime.now(),
            wind_speed=SpeedProbability.example(),
            flood_velocity=SpeedProbability.example(),
            flood_depth=DistanceProbability.example(),
            fire_boundary_dist=DistanceProbability.example(),
            peak_ground_velocity=SpeedProbability.example(),
            peak_ground_acceleration=AccelerationProbability.example(),
        )


class Asset(Component):
    distribution_asset: UUID = Field(..., description="UUID of the distribution asset")
    connections: list[UUID] = Field([], description="List of UUIDs of connected assets")
    devices: list[UUID] = Field(
        [], description="List of UUIDs of devices associated with the asset"
    )
    asset_type: AssetTypes = Field(..., description="Type of the asset")
    height: Distance = Field(..., description="Height of the asset")
    latitude: float = Field(..., ge=-90, le=90, description="Latitude in degrees")
    longitude: float = Field(..., ge=-180, le=180, description="Longitude in degrees")
    asset_state: list[AssetState] = Field(
        ..., description="List of asset states associated with the asset"
    )
    _raster_handler: str | None = None

    @computed_field
    @property
    def elevation(self) -> Distance:
        try:
            elev = Distance(get_elevation(lat=self.latitude, lon=self.longitude), "meter")
        except Exception:
            logger.warning(
                f"Error getting elevation information for asset {self.name} with coordinates {self.latitude}, {self.longitude}. Defaulting to 999 meters"
            )
            elev = Distance(999, "meter")
        return elev

    def _get_asset_state_at_timestamp(self, timestamp: datetime) -> AssetState | None:
        for asset_state in self.asset_state:
            if asset_state.timestamp == timestamp:
                return asset_state
        return AssetState(
            timestamp=timestamp,
        )

    def update_survival_probability(
        self, time_stamp: datetime, hazard_model: hz.BaseDisasterModel, frag_curves
    ):
        asset_state = self._get_asset_state_at_timestamp(time_stamp)
        asset_location = Point(self.longitude, self.latitude)

        if isinstance(hazard_model, hz.EarthQuakeModel):
            asset_state.calculate_earthquake_vectors(asset_location, hazard_model)
        elif isinstance(hazard_model, hz.FireModel):
            asset_state.calculate_fire_vectors(asset_location, hazard_model)
        elif isinstance(hazard_model, hz.WindModel):
            asset_state.calculate_wind_vectors(asset_location, hazard_model)
        elif isinstance(hazard_model, hz.FloodModel):
            asset_state.calculate_flood_vectors(
                asset_location, hazard_model, self.elevation + self.height
            )
        else:
            raise (f"Unsupported hazard type {hazard_model.__class__.__name__}")

        self.calculate_probabilities(asset_state, frag_curves)
        if asset_state not in self.asset_state:
            self.asset_state.append(asset_state)
        return asset_state

    def calculate_probabilities(self, asset_state: AssetState, frag_curves):
        fields = [
            model_field
            for model_field in list(asset_state.model_fields.keys())
            if model_field not in ["uuid", "name", "timestamp"]
        ]
        for field in fields:
            prob_model = getattr(asset_state, field)
            if prob_model and prob_model.survival_probability == 1:
                curve = self.get_vadid_curve(frag_curves, field)
                if curve is None:
                    raise Exception(f"No fragility curve found for field - {field}")
                prob_inst = curve.prob_model
                quantity_name = [
                    model_field
                    for model_field in list(prob_model.model_fields.keys())
                    if model_field not in ["uuid", "name", "survival_probability"]
                ][0]
                quantity = getattr(prob_model, quantity_name)
                prob_model.survival_probability = 1 - prob_inst.probability(quantity)

    def get_vadid_curve(self, frag_curves, field: SyntaxError) -> float:
        for frag_curve in frag_curves:
            if frag_curve.asset_state_param == field:
                for curve in frag_curve.curves:
                    if curve.asset_type == self.asset_type:
                        return curve.prob_function

        return None

    @classmethod
    def example(cls) -> "Asset":
        return Asset(
            name="Asset 1",
            asset_type=AssetTypes.distribution_poles,
            distribution_asset=UUID("123e4567-e89b-12d3-a456-426614174000"),
            height=Distance(100, "m"),
            latitude=37.7749,
            longitude=-122.4194,
            asset_state=[AssetState.example()],
        )



from datetime import datetime

from gdm.distribution import DistributionSystem
from loguru import logger
import numpy as np

from gdm.tracked_changes import (
    TrackedChange,
    PropertyEdit,
)

from erad.default_fragility_curves import DEFAULT_FRAGILTY_CURVES
from erad.models.fragility_curve import HazardFragilityCurves
from erad.systems.hazard_system import HazardSystem
from erad.systems.asset_system import AssetSystem
from erad.constants import HAZARD_TYPES
from erad.models.asset import Asset


class HarzardSimulator:
    def __init__(self, asset_system: AssetSystem):
        self._asset_system = asset_system
        self._asset_system.auto_add_composed_components = True
        self.assets: list[Asset] = list(asset_system.get_components(Asset))

    @classmethod
    def from_gdm(cls, dist_system: DistributionSystem) -> "HarzardSimulator":
        """Create a HarzardSimulator from a DistributionSystem."""
        asset_system = AssetSystem.from_gdm(dist_system)
        return cls(asset_system)

    @property
    def asset_system(self) -> AssetSystem:
        """Get the AssetSystem."""
        return self._asset_system

    def _get_time_stamps(self) -> list[datetime]:
        timestamps = []
        for model_type in HAZARD_TYPES:
            for model in self.hazard_system.get_components(model_type):
                timestamps.append(model.timestamp)
        return sorted(timestamps)

    def run(self, hazard_system: HazardSystem, curve_set: str = "DEFAULT_CURVES"):
        probability_models = list(
            hazard_system.get_components(
                HazardFragilityCurves, filter_func=lambda x: x.name == curve_set
            )
        )

        if not probability_models:
            logger.warning(
                "No HazardFragilityCurves definations found in the passed HazardSystem using default curve definations"
            )
            probability_models = DEFAULT_FRAGILTY_CURVES

        self.hazard_system = hazard_system
        self.timestamps = self._get_time_stamps()
        for timestamp in self.timestamps:
            for hazard_type in HAZARD_TYPES:
                for hazard_model in self.hazard_system.get_components(
                    hazard_type, filter_func=lambda x: x.timestamp == timestamp
                ):
                    for asset in self.assets:
                        assset_state = asset.update_survival_probability(
                            timestamp, hazard_model, probability_models
                        )
                        if not self._asset_system.has_component(assset_state):
                            self._asset_system.add_component(assset_state)


class HazardScenarioGenerator:
    def __init__(
        self,
        asset_system: AssetSystem,
        hazard_system: HazardSystem,
        curve_set: str = "DEFAULT_CURVES",
    ):
        self.assets = list(asset_system.get_components(Asset))
        self.harzard_simulator = HarzardSimulator(asset_system)
        self.harzard_simulator.run(hazard_system, curve_set)

    def _sample(self, scenario_name: str) -> list[TrackedChange]:
        outaged_assets = []
        tracked_changes = []

        n_assets = len(self.assets)
        n_timestamps = len(self.assets[0].asset_state)

        ramdom_samples = np.random.random((n_assets, n_timestamps))

        for ii, asset in enumerate(self.assets):
            for jj, state in enumerate(
                sorted(asset.asset_state, key=lambda asset_state: asset_state.timestamp)
            ):
                if (
                    ramdom_samples[ii, jj] > state.survival_probability
                    and asset.name not in outaged_assets
                ):
                    tracked_changes.append(
                        TrackedChange(
                            scenario_name=scenario_name,
                            timestamp=state.timestamp,
                            edits=[
                                PropertyEdit(
                                    component_uuid=asset.distribution_asset,
                                    name="in_service",
                                    value=False,
                                )
                            ],
                        ),
                    )
                    outaged_assets.append(asset.name)

        return tracked_changes

    def samples(self, number_of_samples: int = 1, seed: int = 0) -> list[TrackedChange]:
        if number_of_samples < 1:
            raise ValueError("number_of_samples should be a positive integer")
        np.random.seed(seed)
        tracked_changes = []
        for i in range(number_of_samples):
            scenario_name = f"sample_{i}"
            tracked_changes.extend(self._sample(scenario_name))
        return tracked_changes



from erad.default_fragility_curves.default_peak_ground_acceleration import (
    DEFAULT_PEAK_GROUND_ACCELERATION_FRAGILITY_CURVES,
)
from erad.default_fragility_curves.default_peak_ground_velocity import (
    DEFAULT_PEAK_GROUND_VELOCITY_FRAGILITY_CURVES,
)
from erad.default_fragility_curves.default_fire_boundary_dist import (
    DEFAULT_FIRE_BOUNDARY_FRAGILITY_CURVES,
)
from erad.default_fragility_curves.default_flood_velocity import (
    DEFAULT_FLOOD_VELOCITY_FRAGILITY_CURVES,
)
from erad.default_fragility_curves.default_flood_depth import DEFAULT_FLOOD_DEPTH_FRAGILITY_CURVES
from erad.default_fragility_curves.default_wind_speed import DEFAULT_WIND_SPEED_FRAGILITY_CURVES
from erad.default_fragility_curves.default_fragility_curves import DEFAULT_FRAGILTY_CURVES



from collections import defaultdict, Counter
from pathlib import Path

from gdm.distribution.enums import PlotingStyle, MapType
from gdm.distribution import DistributionSystem
from shapely.geometry import Point, LineString
import gdm.distribution.components as gdc
from gdm.quantities import Distance
import plotly.graph_objects as go
from infrasys import System
from loguru import logger
import geopandas as gpd
import networkx as nx
import pandas as pd
import numpy as np
import elevation


from erad.constants import ASSET_TYPES, DEFAULT_TIME_STAMP, RASTER_DOWNLOAD_PATH, DEFAULT_HEIGHTS_M
from erad.gdm_mapping import asset_to_gdm_mapping
from erad.models.asset import Asset, AssetState
from erad.enums import AssetTypes, NodeTypes


class AssetSystem(System):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def add_component(self, component, **kwargs):
        assert isinstance(
            component, ASSET_TYPES
        ), f"Unsupported model type {component.__class__.__name__}"
        return super().add_component(component, **kwargs)

    def add_components(self, *components, **kwargs):
        assert all(isinstance(component, ASSET_TYPES) for component in components), (
            "Unsupported model types in passed component. Valid types are: \n"
            + "\n".join([s.__name__ for s in ASSET_TYPES])
        )
        return super().add_components(*components, **kwargs)

    def get_dircted_graph(self):
        """Get the directed graph of the AssetSystem."""

        graph = self.get_undirected_graph()
        substations = list(
            self.get_components(Asset, filter_func=lambda x: x.asset_type == AssetTypes.substation)
        )

        assert len(substations) <= 1, "There should be at most one substation in the asset system."

        tree = nx.dfs_tree(graph, str(substations[0].connections[0]))
        return tree

    def get_undirected_graph(self):
        """Get the undirected graph of the AssetSystem."""
        g = nx.Graph()
        graph_data = []
        for asset in self.get_components(Asset):
            if len(asset.connections) == 2:
                u, v = asset.connections
                g.add_edge(str(u), str(v), **asset.model_dump())
            else:
                if asset.asset_type in NodeTypes:
                    g.add_node(str(asset.distribution_asset), **asset.model_dump())
                else:
                    graph_data.append(asset.model_dump())

        g.graph["metadata"] = graph_data
        return g

    @classmethod
    def from_gdm(cls, dist_system: DistributionSystem) -> "AssetSystem":
        """Create a AssetSystem from a DistributionSystem."""
        asset_map = AssetSystem.map_asets(dist_system)
        # list_of_assets = AssetSystem._build_assets(asset_map)
        system = AssetSystem(auto_add_composed_components=True)
        list_of_assets = system._build_assets(asset_map)
        system.add_components(*list_of_assets)
        return system

    def _add_node_data(
        self, node_data: dict[str, list], asset: Asset, asset_state: AssetState | None = None
    ):
        node_data["name"].append(asset.name)
        node_data["type"].append(asset.asset_type.name)
        node_data["height"].append(asset.height.to("meter").magnitude)
        node_data["elevation"].append(asset.elevation.to("meter").magnitude)
        node_data["latitude"].append(asset.latitude)
        node_data["longitude"].append(asset.longitude)
        node_data["timestamp"].append(asset_state.timestamp if asset_state else DEFAULT_TIME_STAMP)
        node_data["survival_prob"].append(asset_state.survival_probability if asset_state else 1.0)
        node_data["wind_speed"].append(asset_state.wind_speed if asset_state else None)
        node_data["fire_boundary_dist"].append(
            asset_state.fire_boundary_dist if asset_state else None
        )
        node_data["flood_depth"].append(asset_state.flood_depth if asset_state else None)
        node_data["flood_velocity"].append(asset_state.flood_velocity if asset_state else None)
        node_data["peak_ground_acceleration"].append(
            asset_state.peak_ground_acceleration if asset_state else None
        )
        node_data["peak_ground_velocity"].append(
            asset_state.peak_ground_velocity if asset_state else None
        )
        return node_data

    def _add_edge_data(
        self, edge_data: dict[str, list], asset: Asset, asset_state: AssetState | None = None
    ):
        u, v = set(asset.connections)
        buses = list(
            self.get_components(Asset, filter_func=lambda x: x.distribution_asset in [u, v])
        )
        edge_data["name"].append(asset.name)
        edge_data["type"].append(asset.asset_type.name)
        edge_data["height"].append(asset.height.to("meter").magnitude)
        edge_data["elevation"].append(asset.elevation.to("meter").magnitude)
        edge_data["latitude"].append([b.latitude for b in buses])
        edge_data["longitude"].append([b.longitude for b in buses])
        edge_data["timestamp"].append(asset_state.timestamp if asset_state else DEFAULT_TIME_STAMP)
        edge_data["survival_prob"].append(asset_state.survival_probability if asset_state else 1.0)
        edge_data["wind_speed"].append(asset_state.wind_speed if asset_state else None)
        edge_data["fire_boundary_dist"].append(
            asset_state.fire_boundary_dist if asset_state else None
        )
        edge_data["flood_depth"].append(asset_state.flood_depth if asset_state else None)
        edge_data["flood_velocity"].append(asset_state.flood_velocity if asset_state else None)
        edge_data["peak_ground_acceleration"].append(
            asset_state.peak_ground_acceleration if asset_state else None
        )
        edge_data["peak_ground_velocity"].append(
            asset_state.peak_ground_velocity if asset_state else None
        )
        return edge_data

    def to_gdf(self):
        node_data = defaultdict(list)
        edge_data = defaultdict(list)
        assets: list[Asset] = self.get_components(Asset)

        for asset in assets:
            if len(set(asset.connections)) < 2:
                if asset.asset_state:
                    for asset_state in asset.asset_state:
                        node_data = self._add_node_data(node_data, asset, asset_state)
                else:
                    node_data = self._add_node_data(node_data, asset, None)
            else:
                if asset.asset_state:
                    for asset_state in asset.asset_state:
                        edge_data = self._add_edge_data(edge_data, asset, asset_state)
                else:
                    edge_data = self._add_edge_data(edge_data, asset, None)

        nodes_df = pd.DataFrame(node_data)
        gdf_nodes = gpd.GeoDataFrame(
            nodes_df,
            geometry=gpd.points_from_xy(nodes_df.longitude, nodes_df.latitude),
            crs="EPSG:4326",
        )
        edge_df = pd.DataFrame(edge_data)
        edge_df = edge_df[edge_df["longitude"].apply(lambda x: len(x) == 2)]
        geometry = [
            LineString([Point(xy) for xy in zip(*xys)])
            for xys in zip(edge_df["longitude"], edge_df["latitude"])
        ]
        gdf_edges = gpd.GeoDataFrame(edge_df, geometry=geometry, crs="EPSG:4326")
        complete_gdf = pd.concat([gdf_nodes, gdf_edges])
        return complete_gdf

    def to_geojson(self) -> str:
        """Create a GeoJSON from an AssetSystem."""
        gdf_nodes = self.to_gdf()
        return gdf_nodes.to_json()

    def _prepopulate_bus_assets(
        self,
        asset_map: dict[AssetTypes : list[gdc.DistributionComponentBase]],
        list_of_assets: dict[str, Asset],
    ):
        for asset_type, components in asset_map.items():
            for component in components:
                lat, long = AssetSystem._get_component_coordinate(component)
                if isinstance(component, gdc.DistributionBus):
                    list_of_assets[str(component.uuid)] = Asset(
                        name=component.name,
                        connections=[],
                        asset_type=asset_type,
                        distribution_asset=component.uuid,
                        height=Distance(DEFAULT_HEIGHTS_M[asset_type], "meter"),
                        latitude=lat,
                        longitude=long,
                        asset_state=[],
                    )
        return list_of_assets

    def _build_assets(
        self,
        asset_map: dict[AssetTypes : list[gdc.DistributionComponentBase]],
    ) -> list[Asset]:
        list_of_assets: dict[str, Asset] = {}

        self._prepopulate_bus_assets(asset_map, list_of_assets)

        for asset_type, components in asset_map.items():
            for component in components:
                if not isinstance(component, gdc.DistributionBus):
                    lat, long = AssetSystem._get_component_coordinate(component)
                    if hasattr(component, "buses"):
                        connections = [c.uuid for c in component.buses]
                    elif hasattr(component, "bus"):
                        connections = [component.bus.uuid]
                    else:
                        connections = []
                    list_of_assets[str(component.uuid)] = Asset(
                        name=component.name,
                        connections=connections,
                        asset_type=asset_type,
                        distribution_asset=component.uuid,
                        height=Distance(DEFAULT_HEIGHTS_M[asset_type], "meter"),
                        latitude=lat,
                        longitude=long,
                        asset_state=[],
                    )

                    if len(connections) == 1:
                        if str(connections[0]) in list_of_assets:
                            asset = list_of_assets[str(connections[0])]
                            asset.devices.append(component.uuid)

        return list_of_assets.values()

    @staticmethod
    def _get_component_coordinate(component: gdc.DistributionComponentBase):
        if hasattr(component, "buses"):
            xs = [bus.coordinate.x for bus in component.buses]
            ys = [bus.coordinate.y for bus in component.buses]
            if 0 not in xs and 0 not in ys:
                return (sum(xs) / len(xs), sum(ys) / len(ys))
            else:
                return (0, 0)
        elif hasattr(component, "bus"):
            return (component.bus.coordinate.x, component.bus.coordinate.y)
        elif isinstance(component, gdc.DistributionBus):
            return (component.coordinate.x, component.coordinate.y)

    @staticmethod
    def map_asets(
        dist_system: DistributionSystem,
    ) -> dict[AssetTypes : list[gdc.DistributionComponentBase]]:
        asset_dict = defaultdict(list)
        for asset, filters in asset_to_gdm_mapping.items():
            for filter_info in filters:
                models = dist_system.get_components(
                    filter_info.component_type, filter_func=filter_info.component_filter
                )
                asset_dict[asset].extend(list(models))

        AssetSystem._maps_buses(asset_dict, dist_system)
        AssetSystem._map_transformers(asset_dict, dist_system)
        return asset_dict

    @staticmethod
    def _maps_buses(
        asset_dict: dict[AssetTypes : list[gdc.DistributionComponentBase]],
        dist_system: DistributionSystem,
    ):
        for bus in dist_system.get_components(gdc.DistributionBus):
            asset_type = AssetSystem._get_bus_type(bus, asset_dict, dist_system)
            if asset_type:
                asset_dict[asset_type].append(bus)

    @staticmethod
    def _map_transformers(
        asset_dict: dict[AssetTypes : list[gdc.DistributionComponentBase]],
        dist_system: DistributionSystem,
    ):
        for transformer in dist_system.get_components(gdc.DistributionTransformerBase):
            bus_types = [
                AssetSystem._get_bus_type(b, asset_dict, dist_system) for b in transformer.buses
            ]
            if (
                AssetTypes.transmission_junction_box in bus_types
                or AssetTypes.transmission_tower in bus_types
            ):
                asset_dict[AssetTypes.transformer_mad_mount].append(transformer)
            elif all(bus_type == AssetTypes.distribution_junction_box for bus_type in bus_types):
                asset_dict[AssetTypes.transformer_mad_mount].append(transformer)
            else:
                asset_dict[AssetTypes.transformer_mad_mount].append(transformer)

    @staticmethod
    def _get_bus_type(
        bus: gdc.DistributionBus,
        asset_dict: dict[AssetTypes : list[gdc.DistributionComponentBase]],
        dist_system: DistributionSystem,
    ):
        components = dist_system.get_bus_connected_components(bus.name, gdc.DistributionBranchBase)
        connected_types = []
        for component in components:
            if component in asset_dict[AssetTypes.distribution_overhead_lines]:
                connected_types.append(AssetTypes.distribution_poles)
            elif component in asset_dict[AssetTypes.transmission_overhead_lines]:
                connected_types.append(AssetTypes.transmission_tower)
            elif component in asset_dict[AssetTypes.transmission_underground_cables]:
                connected_types.append(AssetTypes.transmission_junction_box)
            elif component in asset_dict[AssetTypes.distribution_underground_cables]:
                connected_types.append(AssetTypes.distribution_junction_box)
            else:
                ...
        if connected_types:
            counter = Counter(connected_types)
            return counter.most_common(1)[0][0]

    def _add_node_traces(
        self,
        time_index: int,
        fig: go.Figure,
        df_ts: gpd.GeoDataFrame,
        plotting_object: go.Scattermap | go.Scattergeo,
    ):
        points_only = df_ts[df_ts.geometry.geom_type == "Point"]
        for asset_type in set(points_only["type"]):
            df_filt = points_only[points_only["type"] == asset_type]
            text = [
                "<br>".join([f"<b>{kk}:</b> {vv}" for kk, vv in rr.to_dict().items()][:-1])
                for __, rr in df_filt.iterrows()
            ]
            trace = plotting_object(
                lat=df_filt.geometry.y,
                lon=df_filt.geometry.x,
                mode="markers",
                marker=dict(size=10),
                name=asset_type,
                hovertext=text,
                visible=(time_index == 0),
            )
            fig.add_trace(trace)
        return fig

    def _add_edge_traces(
        self,
        time_index: int,
        fig: go.Figure,
        df_ts: gpd.GeoDataFrame,
        plotting_object: go.Scattermap | go.Scattergeo,
    ):
        points_only = df_ts[df_ts.geometry.geom_type == "LineString"]
        for asset_type in set(points_only["type"]):
            df_filt = points_only[points_only["type"] == asset_type]
            features = {k: [] for k in list(df_filt.columns) + ["text"]}
            for _, row in df_filt.iterrows():
                for c in df_filt.columns:
                    features[c] = np.append(features[c], row[c])
                features["text"] = np.append(
                    features["text"],
                    "<br> ".join([f"<b>{kk}:</b> {vv}" for kk, vv in row.to_dict().items()][:-1]),
                )
                for c in df_filt.columns:
                    features[c] = np.append(features[c], None)
                features["text"] = np.append(features["text"], None)

            trace = plotting_object(
                name=asset_type,
                lat=features["latitude"],
                lon=features["longitude"],
                mode="lines",
                hovertext=features["text"],
                visible=(time_index == 0),
            )
            fig.add_trace(trace)
        return fig

    def _has_zero_zero_coords(self, geom):
        if geom.is_empty or geom is None:
            return False
        if geom.geom_type == "Point":
            return geom.x == 0 and geom.y == 0
        elif geom.geom_type in ["Polygon", "MultiPolygon"]:
            return any(x == 0 and y == 0 for x, y in geom.exterior.coords)
        elif geom.geom_type == "LineString":
            return any(x == 0 and y == 0 for x, y in geom.coords)
        elif geom.geom_type.startswith("Multi") or geom.geom_type == "GeometryCollection":
            return any(self._has_zero_zero_coords(g) for g in geom.geoms)
        return False

    def get_elevation_raster(self) -> Path:
        """Download and clip elevation raster for the AssetSystem."""
        if RASTER_DOWNLOAD_PATH.exists():
            RASTER_DOWNLOAD_PATH.unlink()

        coordinates = np.array(
            [
                (asset.latitude, asset.longitude)
                for asset in self.get_components(Asset)
                if asset.latitude and asset.latitude
            ]
        )

        if len(coordinates):
            lat_min, lat_max = float(coordinates[:, 0].min()), float(coordinates[:, 0].max())
            lon_min, lon_max = float(coordinates[:, 1].min()), float(coordinates[:, 1].max())
            bounds = (lon_min, lat_min, lon_max, lat_max)
            logger.info(f"Downloading raster file to path: {RASTER_DOWNLOAD_PATH}")
            logger.info(f"Clipping raster for bounds: {bounds}")
            elevation.clip(bounds=bounds, output=str(RASTER_DOWNLOAD_PATH.name))
            if not RASTER_DOWNLOAD_PATH.exists():
                raise FileNotFoundError(f"File path {RASTER_DOWNLOAD_PATH} does not exist")
        else:
            logger.info(
                "No assets found in the AssetSystem, no elevation raster will be downloaded."
            )
            return None

    def plot(
        self,
        show: bool = True,
        show_legend: bool = True,
        map_type: MapType = MapType.SCATTER_MAP,
        style: PlotingStyle = PlotingStyle.CARTO_POSITRON,
        zoom_level: int = 11,
        figure=go.Figure(),
    ):
        """Plot the AssetSystem."""
        plotting_object = getattr(go, map_type.value)
        gdf = self.to_gdf()
        gdf["timestamp"] = pd.to_datetime(gdf["timestamp"])
        gdf = gdf[~gdf.geometry.apply(self._has_zero_zero_coords)]
        timestamps = sorted(gdf["timestamp"].unique())

        steps = []

        for i, ts in enumerate(timestamps):
            df_ts = gdf[gdf["timestamp"] == ts]

            figure = self._add_node_traces(i, figure, df_ts, plotting_object)
            figure = self._add_edge_traces(i, figure, df_ts, plotting_object)

            steps.append(
                dict(
                    method="update",
                    label=str(ts.date()),
                    args=[{"visible": [j == i for j in range(len(timestamps))]}],
                )
            )

        sliders = [dict(active=0, pad={"t": 50}, steps=steps)]

        figure.update_layout(
            mapbox=dict(
                style=style.value,
                zoom=zoom_level,
                # center=dict(lat=gdf.geometry.y.mean(), lon=gdf.geometry.x.mean()),
            ),
            sliders=sliders,
            showlegend=True if show_legend else False,
        )

        if show:
            figure.show()

        return figure
